"""
Tools for integrating `wandb` with [`Keras`](https://keras.io/), a deep learning API for [`TensorFlow`](https://www.tensorflow.org/).

Use the `WandbCallback` to add `wandb` logging to any `Keras` model.
"""

from .keras import WandbCallback

__all__ = ["WandbCallback"]
